<?php
/**
 * Copyright (C) 2014-2017 ServMask Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ███████╗███████╗██████╗ ██╗   ██╗███╗   ███╗ █████╗ ███████╗██╗  ██╗
 * ██╔════╝██╔════╝██╔══██╗██║   ██║████╗ ████║██╔══██╗██╔════╝██║ ██╔╝
 * ███████╗█████╗  ██████╔╝██║   ██║██╔████╔██║███████║███████╗█████╔╝
 * ╚════██║██╔══╝  ██╔══██╗╚██╗ ██╔╝██║╚██╔╝██║██╔══██║╚════██║██╔═██╗
 * ███████║███████╗██║  ██║ ╚████╔╝ ██║ ╚═╝ ██║██║  ██║███████║██║  ██╗
 * ╚══════╝╚══════╝╚═╝  ╚═╝  ╚═══╝  ╚═╝     ╚═╝╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝
 */

class ServMaskFTPClient
{
	/**
	 * Holds FTP client instance
	 *
	 * @access protected
	 * @var    object $client FTP client instance
	 */
	protected $client    = null;

	/**
	 * Holds FTP server hostname
	 *
	 * @access protected
	 * @var    string $hostname FTP server hostname
	 */
	protected $hostname  = null;

	/**
	 * Holds FTP client username
	 *
	 * @access protected
	 * @var    string $username FTP client username
	 */
	protected $username  = null;

	/**
	 * Holds FTP client password
	 *
	 * @access protected
	 * @var    string $password FTP client password
	 */
	protected $password  = null;

	/**
	 * Holds FTP server directory
	 *
	 * @access protected
	 * @var    string $directory FTP server directory
	 */
	protected $directory = null;

	/**
	 * Holds FTP server port
	 *
	 * @access protected
	 * @var    integer $port FTP server port
	 */
	protected $port      = null;

	/**
	 * Holds FTP pasive mode
	 *
	 * @access protected
	 * @var    boolean $passive FTP pasive mode
	 */
	protected $passive   = null;

	/**
	 * Holds FTP SSL mode
	 *
	 * @access protected
	 * @var    boolean $ssl FTP SSL mode
	 */
	protected $ssl       = null;

	/**
	 * Constructs the ServMaskFTPClient object
	 *
	 * @param string  $hostname  FTP server hostname
	 * @param string  $username  FTP client username
	 * @param string  $password  FTP client password
	 * @param string  $directory FTP server directory
	 * @param integer $port      FTP server port
	 * @param boolean $passive   FTP passive mode
	 * @param boolean $ssl       FTP SSL mode
	 */
	public function __construct( $hostname, $username, $password, $directory = '/', $port = 21, $passive = false, $ssl = false ) {
		$this->hostname  = $hostname;
		$this->username  = $username;
		$this->password  = $password;
		$this->directory = $directory;
		$this->port      = $port;
		$this->passive   = $passive;
		$this->ssl       = $ssl;
	}

	/**
	 * Create folder
	 *
	 * @param  string  $path FTP path
	 * @return boolean
	 */
	public function createFolder( $path ) {
		return $this->getClient()->createFolder( $path );
	}

	/**
	 * Retrieve files and folders metadata
	 *
	 * @param  string $path FTP path
	 * @return array
	 */
	public function listFolder( $path ) {
		$response = array();

		// Get files and folders
		$items = $this->getClient()->listFolder( $path );

		// Loop over files and folders
		foreach ( $items as $item ) {
			$response[] = array(
				'name'  => isset( $item['name'] ) ? $item['name'] : null,
				'path'  => isset( $item['path'] ) ? $item['path'] : null,
				'date'  => isset( $item['date'] ) ? $item['date'] : null,
				'size'  => isset( $item['size'] ) ? size_format( $item['size'] ) : null,
				'bytes' => isset( $item['size'] ) ? $item['size'] : null,
				'type'  => isset( $item['dir'] ) && $item['dir'] === true ? 'folder' : 'file',
			);
		}

		return $response;
	}

	/**
	 * Upload file
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function uploadFile( $remoteFile, $localFile, &$params = array() ) {
		return $this->getClient()->uploadFile( $remoteFile, $localFile, $params );
	}

	/**
	 * Upload file chunk
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function uploadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		return $this->getClient()->uploadFileChunk( $remoteFile, $localFile, $params );
	}

	/**
	 * Download file
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function downloadFile( $remoteFile, $localFile, &$params = array() ) {
		return $this->getClient()->downloadFile( $remoteFile, $localFile, $params );
	}

	/**
	 * Download file chunk
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function downloadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		return $this->getClient()->downloadFileChunk( $remoteFile, $localFile, $params );
	}

	/**
	 * Remove file
	 *
	 * @param  string  $path FTP path
	 * @return boolean
	 */
	public function removeFile( $path ) {
		return $this->getClient()->removeFile( $path );
	}

	/**
	 * Test FTP connection
	 *
	 * @return boolean
	 */
	public function testConnection() {
		return $this->getClient()->testConnection();
	}

	/**
	 * Test FTP append file
	 *
	 * @param  string $remoteFile Path to remote file
	 * @return boolean
	 */
	public function testAppendFile( $remoteFile = '.servmask' ) {
		return $this->getClient()->testAppendFile( $remoteFile );
	}

	/**
	 * Get FTP client (lazy loading)
	 *
	 * @return ServMaskFTPInterface
	 */
	protected function getClient() {
		if ( $this->client === null ) {
			try {
				$this->client = new ServMaskFTPCurl( $this->ssl );
			} catch ( Exception $e ) {
				$this->client = new ServMaskFTPExtension( $this->ssl );
			}

			// Set configuration
			$this->client->setHostname( $this->hostname );
			$this->client->setUsername( $this->username );
			$this->client->setPassword( $this->password );
			$this->client->setDirectory( $this->directory );
			$this->client->setPort( $this->port );
			$this->client->setPassive( $this->passive );
		}

		return $this->client;
	}
}
