<?php
/**
 * Copyright (C) 2014-2017 ServMask Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ███████╗███████╗██████╗ ██╗   ██╗███╗   ███╗ █████╗ ███████╗██╗  ██╗
 * ██╔════╝██╔════╝██╔══██╗██║   ██║████╗ ████║██╔══██╗██╔════╝██║ ██╔╝
 * ███████╗█████╗  ██████╔╝██║   ██║██╔████╔██║███████║███████╗█████╔╝
 * ╚════██║██╔══╝  ██╔══██╗╚██╗ ██╔╝██║╚██╔╝██║██╔══██║╚════██║██╔═██╗
 * ███████║███████╗██║  ██║ ╚████╔╝ ██║ ╚═╝ ██║██║  ██║███████║██║  ██╗
 * ╚══════╝╚══════╝╚═╝  ╚═╝  ╚═══╝  ╚═╝     ╚═╝╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝
 */

require_once dirname( __FILE__ ) . DIRECTORY_SEPARATOR . 'ServMaskFTPInterface.php';

class ServMaskFTPCurl implements ServMaskFTPInterface
{
	const CHUNK_SIZE     = 4194304; // 4 MB

	/**
	 * Holds FTP server hostname
	 *
	 * @access protected
	 * @var    string $hostname FTP server hostname
	 */
	protected $hostname  = null;

	/**
	 * Holds FTP client username
	 *
	 * @access protected
	 * @var    string $username FTP client username
	 */
	protected $username  = null;

	/**
	 * Holds FTP client password
	 *
	 * @access protected
	 * @var    string $password FTP client password
	 */
	protected $password  = null;

	/**
	 * Holds FTP server directory
	 *
	 * @access protected
	 * @var    string $directory FTP server directory
	 */
	protected $directory = null;

	/**
	 * Holds FTP server port
	 *
	 * @access protected
	 * @var    integer $port FTP server port
	 */
	protected $port      = null;

	/**
	 * Holds FTP pasive mode
	 *
	 * @access protected
	 * @var    boolean $passive FTP pasive mode
	 */
	protected $passive   = null;

	/**
	 * Holds FTP SSL mode
	 *
	 * @access protected
	 * @var    boolean $ssl FTP SSL mode
	 */
	protected $ssl       = null;

	/**
	 * Constructs the ServMaskFTPCurl object
	 *
	 * @param boolean $ssl FTP SSL mode
	 */
	public function __construct( $ssl = false ) {
		// Check the cURL extension is loaded
		if ( ! extension_loaded( 'curl' ) ) {
			throw new Exception( 'cURL extension is required.' );
		}

		$this->ssl = $ssl;
	}

	/**
	 * Set hostname
	 *
	 * @param  string          $value FTP hostname
	 * @return ServMaskFTPCurl
	 */
	public function setHostname( $value ) {
		$this->hostname = $value;
		return $this;
	}

	/**
	 * Get hostname
	 *
	 * @return string
	 */
	public function getHostname() {
		return $this->hostname;
	}

	/**
	 * Set username
	 *
	 * @param  string          $username FTP Username
	 * @return ServMaskFTPCurl
	 */
	public function setUsername( $value ) {
		$this->username = $value;
		return $this;
	}

	/**
	 * Get username
	 *
	 * @return string
	 */
	public function getUsername() {
		return $this->username;
	}

	/**
	 * Set password
	 *
	 * @param  string          $value FTP Password
	 * @return ServMaskFTPCurl
	 */
	public function setPassword( $value ) {
		$this->password = $value;
		return $this;
	}

	/**
	 * Get password
	 *
	 * @return string
	 */
	public function getPassword() {
		return $this->password;
	}

	/**
	 * Set directory
	 *
	 * @param  string          $value FTP Directory
	 * @return ServMaskFTPCurl
	 */
	public function setDirectory( $value ) {
		$this->directory = $value;
		return $this;
	}

	/**
	 * Get directory
	 *
	 * @return string
	 */
	public function getDirectory() {
		return $this->directory;
	}

	/**
	 * Set port
	 *
	 * @param  integer         $value FTP Port
	 * @return ServMaskFTPCurl
	 */
	public function setPort( $value ) {
		$this->port = intval( $value );
		return $this;
	}

	/**
	 * Get port
	 *
	 * @return integer
	 */
	public function getPort() {
		return $this->port;
	}

	/**
	 * Set passive mode
	 *
	 * @param  boolean         $value FTP Passive Mode
	 * @return ServMaskFTPCurl
	 */
	public function setPassive( $value ) {
		$this->passive = $value;
		return $this;
	}

	/**
	 * Get passive mode
	 *
	 * @return boolean
	 */
	public function getPassive() {
		return $this->passive;
	}

	/**
	 * Create folder
	 *
	 * @param  string  $path FTP path
	 * @return boolean
	 */
	public function createFolder( $path ) {
		return $this->makeRequest( sprintf( '/%s/%s/', $this->getDirectory(), $path ), array(
			CURLOPT_FTP_CREATE_MISSING_DIRS => true,
		) );
	}

	/**
	 * Retrieve files and folders metadata
	 *
	 * @param  string $path FTP path
	 * @return array
	 */
	public function listFolder( $path ) {
		// List raw files and directories
		$rawItems = $this->rawListFolder( $path );

		// List files and directories
		$items = $this->makeRequest( sprintf( '/%s/%s/', $this->getDirectory(), $path ), array(
			CURLOPT_FTPLISTONLY => true,
		) );

		// Loop over files and folders
		$objects = array();
		foreach ( explode( "\n", $items ) as $item ) {
			if ( ( $item = trim( $item ) ) && strpos( $item, '.' ) !== 0 ) {
				foreach ( $rawItems as $rawItem ) {
					if ( ( $rawItem = trim( $rawItem ) ) && strpos( strrev( $rawItem ), strrev( $item ) ) === 0 ) {
						if ( ( $meta = preg_split( '/\s+/', $rawItem ) ) ) {
							$objects[] = array(
								'name' => basename( $item ),
								'path' => $this->sanitizePath( sprintf( '/%s/%s', $path, $item ) ),
								'date' => @strtotime( "{$meta[5]} {$meta[6]} {$meta[7]}" ),
								'size' => @$meta[4],
								'dir'  => @strpos( $rawItem, 'd' ) === 0,
							);
						}
					}
				}
			}
		}

		return $objects;
	}

	/**
	 * Retrieve files and folders raw metadata
	 *
	 * @param  string $path FTP path
	 * @return array
	 */
	public function rawListFolder( $path ) {
		// List files and directories
		$rawItems = $this->makeRequest( sprintf( '/%s/%s/', $this->getDirectory(), $path ), array(
			CURLOPT_CUSTOMREQUEST => 'LIST -tlA',
		) );

		return explode( "\n", $rawItems );
	}

	/**
	 * Upload file
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function uploadFile( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );
		$localFileSize = filesize( $localFileMeta['uri'] );

		// Upload the file
		$this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ), array(
			CURLOPT_UPLOAD     => true,
			CURLOPT_INFILE     => $localFile,
			CURLOPT_INFILESIZE => $localFileSize,
		) );

		// Set offset
		$params['offset'] = $localFileSize;

		return $params;
	}

	/**
	 * Upload file chunk
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function uploadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );
		$localFileSize = filesize( $localFileMeta['uri'] );

		// Get chunk stream
		$chunk = fopen( 'php://temp', 'wb+' );

		// Get chunk size
		if ( $localFileSize < ( $params['offset'] + self::CHUNK_SIZE ) ) {
			$chunkSize = $localFileSize - $params['offset'];
		} else {
			$chunkSize = self::CHUNK_SIZE;
		}

		// Read file chunk
		if ( ( $size = stream_copy_to_stream( $localFile, $chunk, $chunkSize, $params['offset'] ) ) ) {
			// Set chunk pointer
			rewind( $chunk );

			// Upload the chunk
			$this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ), array(
				CURLOPT_UPLOAD     => true,
				CURLOPT_FTPAPPEND  => true,
				CURLOPT_INFILE     => $chunk,
				CURLOPT_INFILESIZE => $size,
			) );

			// Set offset
			$params['offset'] += $size;
		}

		return $params;
	}

	/**
	 * Download file
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function downloadFile( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );

		// Download the file
		$this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ), array(
			CURLOPT_FILE => $localFile,
		) );

		// Set offset
		$params['offset'] = filesize( $localFileMeta['uri'] );

		return $params;
	}

	/**
	 * Download file chunk
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function downloadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		// Download the file
		$this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ), array(
			CURLOPT_FILE => $localFile,
		) );

		$localFileMeta = stream_get_meta_data( $localFile );

		// Set offset
		$params['offset'] = filesize( $localFileMeta['uri'] );

		return $params;
	}

	/**
	 * Remove file
	 *
	 * @param  string  $remoteFile FTP path
	 * @return boolean
	 */
	public function removeFile( $remoteFile ) {
		return $this->makeRequest( sprintf( '/%s/', $this->getDirectory() ), array(
			CURLOPT_QUOTE => array( sprintf( 'DELE /%s', $this->sanitizePath( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ) ) ) ),
		) );
	}

	/**
	 * Test FTP connection
	 *
	 * @return boolean
	 */
	public function testConnection() {
		try {
			return (bool) $this->makeRequest( sprintf( '/%s/', $this->getDirectory() ) );
		} catch ( Exception $e ) {
			return false;
		}
	}

	/**
	 * Test FTP append file
	 *
	 * @param  string $remoteFile Path to remote file
	 * @return boolean
	 */
	public function testAppendFile( $remoteFile = '.servmask' ) {

		try {

			// Write temporary file
			if ( ( $localFile = fopen( 'php://temp', 'wb+' ) ) ) {

				// Write PHP OS
				fwrite( $localFile, PHP_OS );

				// Set file pointer
				rewind( $localFile );

				// Upload file
				$this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ), array(
					CURLOPT_UPLOAD     => true,
					CURLOPT_INFILE     => $localFile,
					CURLOPT_INFILESIZE => $size,
				) );

				// Close file handler
				fclose( $localFile );
			}

			// Write temporary file (append)
			if ( ( $localFile = fopen( 'php://temp', 'wb+' ) ) ) {

				// Write PHP OS
				fwrite( $localFile, PHP_OS );

				// Set file pointer
				rewind( $localFile );

				// Upload file (append)
				$this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ), array(
					CURLOPT_UPLOAD     => true,
					CURLOPT_FTPAPPEND  => true,
					CURLOPT_INFILE     => $localFile,
					CURLOPT_INFILESIZE => $size,
				) );

				// Close file handler
				fclose( $localFile );
			}

			// Get file
			if ( ( $remoteContent = $this->makeRequest( sprintf( '/%s/%s', $this->getDirectory(), $remoteFile ) ) ) ) {
				if ( strlen( $remoteContent ) === ( 2 * strlen( PHP_OS ) ) ) {
					return true;
				}
			}

		} catch ( Exception $e ) {
		}

		return false;
	}

	/**
	 * Make cURL request
	 *
	 * @param  string $path    FTP path
	 * @param  array  $options cURL options
	 * @return mixed
	 */
	protected function makeRequest( $path, $options = array() ) {
		$path = $this->sanitizePath( $path );

		// Open cURL handler
		$handler = curl_init();

		// Set FTP URL
		curl_setopt( $handler, CURLOPT_URL, sprintf( 'ftp://%s:%d/%s', $this->getHostname(), $this->getPort(), $path ) );

		// Set username and password
		curl_setopt( $handler, CURLOPT_USERPWD, sprintf( '%s:%s', $this->getUsername(), $this->getPassword() ) );

		// Set default configuration
		curl_setopt( $handler, CURLOPT_HEADER, false );
		curl_setopt( $handler, CURLOPT_RETURNTRANSFER, true );
		curl_setopt( $handler, CURLOPT_BINARYTRANSFER, true );
		curl_setopt( $handler, CURLOPT_CONNECTTIMEOUT, 120 );
		curl_setopt( $handler, CURLOPT_TIMEOUT, 0 );

		// Add additional options to connect to FTP with SSL if ssl was selected
		if ( $this->ssl ) {
			curl_setopt( $handler, CURLOPT_SSL_VERIFYPEER, false );
			curl_setopt( $handler, CURLOPT_SSL_VERIFYHOST, false );
			curl_setopt( $handler, CURLOPT_FTP_SSL, CURLFTPSSL_TRY );
			curl_setopt( $handler, CURLOPT_FTPSSLAUTH, CURLFTPAUTH_TLS );
		}

		// Turn on Passive Mode
		if ( $this->getPassive() ) {
			curl_setopt( $handler, CURLOPT_FTP_USE_EPSV, true );
		}

		// Apply cURL options
		foreach ( $options as $name => $value ) {
			curl_setopt( $handler, $name, $value );
		}

		// HTTP request
		$response = curl_exec( $handler );
		if ( $response === false ) {
			throw new Exception( sprintf( 'Unable to connect to FTP. Error code: %d', curl_errno( $handler ) ) );
		}

		// Close cURL handler
		curl_close( $handler );

		return $response;
	}

	/**
	 * Sanitize FTP path
	 *
	 * @param  string $path FTP path
	 * @return string
	 */
	protected function sanitizePath( $path ) {
		return ltrim( preg_replace( '/[\\\\\/]+/', '/', $path ), '/' );
	}
}
