<?php
/**
 * Copyright (C) 2014-2017 ServMask Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ███████╗███████╗██████╗ ██╗   ██╗███╗   ███╗ █████╗ ███████╗██╗  ██╗
 * ██╔════╝██╔════╝██╔══██╗██║   ██║████╗ ████║██╔══██╗██╔════╝██║ ██╔╝
 * ███████╗█████╗  ██████╔╝██║   ██║██╔████╔██║███████║███████╗█████╔╝
 * ╚════██║██╔══╝  ██╔══██╗╚██╗ ██╔╝██║╚██╔╝██║██╔══██║╚════██║██╔═██╗
 * ███████║███████╗██║  ██║ ╚████╔╝ ██║ ╚═╝ ██║██║  ██║███████║██║  ██╗
 * ╚══════╝╚══════╝╚═╝  ╚═╝  ╚═══╝  ╚═╝     ╚═╝╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝
 */

require_once dirname( __FILE__ ) . DIRECTORY_SEPARATOR . 'ServMaskFTPInterface.php';

class ServMaskFTPExtension implements ServMaskFTPInterface
{
	/**
	 * Holds FTP server hostname
	 *
	 * @access protected
	 * @var    string $hostname FTP server hostname
	 */
	protected $hostname  = null;

	/**
	 * Holds FTP client username
	 *
	 * @access protected
	 * @var    string $username FTP client username
	 */
	protected $username  = null;

	/**
	 * Holds FTP client password
	 *
	 * @access protected
	 * @var    string $password FTP client password
	 */
	protected $password  = null;

	/**
	 * Holds FTP server directory
	 *
	 * @access protected
	 * @var    string $directory FTP server directory
	 */
	protected $directory = null;

	/**
	 * Holds FTP server port
	 *
	 * @access protected
	 * @var    integer $port FTP server port
	 */
	protected $port      = null;

	/**
	 * Holds FTP pasive mode
	 *
	 * @access protected
	 * @var    boolean $passive FTP pasive mode
	 */
	protected $passive   = null;

	/**
	 * Holds FTP SSL mode
	 *
	 * @access protected
	 * @var    boolean $ssl FTP SSL mode
	 */
	protected $ssl       = null;

	/**
	 * Holds FTP handler
	 *
	 * @access protected
	 * @var    resource $handler FTP handler
	 */
	protected $handler   = null;


	public function __construct( $ssl = false ) {
		// Check the FTP extension is loaded
		if ( ! extension_loaded( 'ftp' ) ) {
			throw new Exception( 'FTP extension is required.' );
		}

		$this->ssl = $ssl;
	}

	/**
	 * Set hostname
	 *
	 * @param  string               $value FTP hostname
	 * @return ServMaskFTPExtension
	 */
	public function setHostname( $value ) {
		$this->hostname = $value;
		return $this;
	}

	/**
	 * Get hostname
	 *
	 * @return string
	 */
	public function getHostname() {
		return $this->hostname;
	}

	/**
	 * Set username
	 *
	 * @param  string               $username FTP Username
	 * @return ServMaskFTPExtension
	 */
	public function setUsername( $value ) {
		$this->username = $value;
		return $this;
	}

	/**
	 * Get username
	 *
	 * @return string
	 */
	public function getUsername() {
		return $this->username;
	}

	/**
	 * Set password
	 *
	 * @param  string               $value FTP Password
	 * @return ServMaskFTPExtension
	 */
	public function setPassword( $value ) {
		$this->password = $value;
		return $this;
	}

	/**
	 * Get password
	 *
	 * @return string
	 */
	public function getPassword() {
		return $this->password;
	}

	/**
	 * Set directory
	 *
	 * @param  string               $value FTP Directory
	 * @return ServMaskFTPExtension
	 */
	public function setDirectory( $value ) {
		$this->directory = $value;
		return $this;
	}

	/**
	 * Get directory
	 *
	 * @return string
	 */
	public function getDirectory() {
		return $this->directory;
	}

	/**
	 * Set port
	 *
	 * @param  integer              $value FTP Port
	 * @return ServMaskFTPExtension
	 */
	public function setPort( $value ) {
		$this->port = intval( $value );
		return $this;
	}

	/**
	 * Get port
	 *
	 * @return integer
	 */
	public function getPort() {
		return $this->port;
	}

	/**
	 * Set passive mode
	 *
	 * @param  boolean              $value FTP Passive Mode
	 * @return ServMaskFTPExtension
	 */
	public function setPassive( $value ) {
		$this->passive = $value;
		return $this;
	}

	/**
	 * Get passive mode
	 *
	 * @return boolean
	 */
	public function getPassive() {
		return $this->passive;
	}

	/**
	 * Create folder
	 *
	 * @param  string  $path FTP path
	 * @return boolean
	 */
	public function createFolder( $path ) {
		$path = $this->sanitizePath( $path );

		// Create new folder
		if ( ! @ftp_nlist( $this->getConnection(), $path ) ) {
			if ( @ftp_mkdir( $this->getConnection(), $path ) ) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Retrieve files and folders metadata
	 *
	 * @param  string $path FTP path
	 * @return array
	 */
	public function listFolder( $path ) {
		$path = $this->sanitizePath( $path );

		// List raw files and directories
		if ( empty( $path ) ) {
			$rawItems = $this->rawListFolder( '.' );
		} else {
			$rawItems = $this->rawListFolder( $path );
		}

		// Loop over files and folders
		$objects = array();
		if ( ( $items = @ftp_nlist( $this->getConnection(), $path ) ) ) {
			foreach ( $items as $item ) {
				if ( ( $item = trim( $item ) ) && strpos( $item, '.' ) !== 0 ) {
					foreach ( $rawItems as $rawItem ) {
						if ( ( $rawItem = trim( $rawItem ) ) && strpos( strrev( $rawItem ), strrev( $item ) ) === 0 ) {
							if ( ( $meta = preg_split( '/\s+/', $rawItem ) ) ) {
								$objects[] = array(
									'name' => basename( $item ),
									'path' => $this->sanitizePath( sprintf( '/%s/%s', $path, $item ) ),
									'date' => @strtotime( "{$meta[5]} {$meta[6]} {$meta[7]}" ),
									'size' => @$meta[4],
									'dir'  => @strpos( $rawItem, 'd' ) === 0,
								);
							}
						}
					}
				}
			}
		}

		return $objects;
	}

	/**
	 * Retrieve files and folders raw metadata
	 *
	 * @param  string $path FTP path
	 * @return array
	 */
	public function rawListFolder( $path ) {
		$path = $this->sanitizePath( $path );

		// Get list of files and directories
		$objects = array();
		if ( ( $items = @ftp_rawlist( $this->getConnection(), sprintf( '-tlA %s', $path ) ) ) ) {
			foreach ( $items as $item ) {
				$objects[] = $item;
			}
		}

		return $objects;
	}

	/**
	 * Upload file
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function uploadFile( $remoteFile, $localFile, &$params = array() ) {
		$remoteFile = $this->sanitizePath( $remoteFile );

		// Get local file meta
		$localFileMeta = stream_get_meta_data( $localFile );

		// Upload the file
		if ( is_resource( $localFile ) ) {
			if ( false === @ftp_fput( $this->getConnection(), $remoteFile, $localFile, FTP_BINARY ) ) {
				throw new Exception( 'FTP upload has failed.' );
			}
		} else {
			if ( false === @ftp_put( $this->getConnection(), $remoteFile, $localFile, FTP_BINARY ) ) {
				throw new Exception( 'FTP upload has failed.' );
			}
		}

		// Set offset
		$params['offset'] = filesize( $localFileMeta['uri'] );

		return $params;
	}

	/**
	 * Upload file chunk
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function uploadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		$remoteFile = $this->sanitizePath( $remoteFile );

		// Get file offset
		$params['offset'] = @ftp_size( $this->getConnection(), $remoteFile );

		// Upload the file
		if ( is_resource( $localFile ) ) {
			$status = @ftp_nb_fput( $this->getConnection(), $remoteFile, $localFile, FTP_BINARY, $params['offset'] );
		} else {
			$status = @ftp_nb_put( $this->getConnection(), $remoteFile, $localFile, FTP_BINARY, $params['offset'] );
		}

		// Check upload status
		if ( $status == FTP_FAILED ) {
			throw new Exception( 'FTP upload has failed.' );
		}

		// Start time
		$start = microtime( true );

		while ( $status == FTP_MOREDATA ) {
			// Continue upload
			$status = @ftp_nb_continue( $this->getConnection() );

			// More than 10 seconds have passed, break and do another request
			if ( ( microtime( true ) - $start ) > 10 ) {
				break;
			}
		}

		return $params;
	}

	/**
	 * Download file
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function downloadFile( $remoteFile, $localFile, &$params = array() ) {
		$remoteFile = $this->sanitizePath( $remoteFile );

		// Set offset
		$params['offset'] = @ftp_size( $this->getConnection(), $remoteFile );

		// Download the file
		if ( is_resource( $localFile ) ) {
			if ( false === @ftp_fget( $this->getConnection(), $localFile, $remoteFile, FTP_BINARY ) ) {
				throw new Exception( 'FTP download has failed.' );
			}
		} else {
			if ( false === @ftp_get( $this->getConnection(), $localFile, $remoteFile, FTP_BINARY ) ) {
				throw new Exception( 'FTP download has failed.' );
			}
		}

		return $params;
	}

	/**
	 * Download file chunk
	 *
	 * @param  string $remoteFile FTP path
	 * @param  mixed  $localFile  File path or file stream
	 * @param  array  $params     File parameters
	 * @return array
	 */
	public function downloadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		$remoteFile = $this->sanitizePath( $remoteFile );

		// Get file size
		$params['size'] = @ftp_size( $this->getConnection(), $remoteFile );

		// Download the file
		if ( is_resource( $localFile ) ) {
			$status = @ftp_nb_fget( $this->getConnection(), $localFile, $remoteFile, FTP_BINARY, $params['offset'] );
		} else {
			$status = @ftp_nb_get( $this->getConnection(), $localFile, $remoteFile, FTP_BINARY, $params['offset'] );
		}

		// Check download status
		if ( $status == FTP_FAILED ) {
			throw new Exception( 'FTP download has failed.' );
		}

		// Start time
		$start = microtime( true );

		while ( $status == FTP_MOREDATA ) {
			// Continue download
			$status = @ftp_nb_continue( $this->getConnection() );

			// More than 10 seconds have passed, break and do another request
			if ( ( microtime( true ) - $start ) > 10 ) {
				break;
			}
		}

		return $params;
	}

	/**
	 * Remove file
	 *
	 * @param  string  $remoteFile FTP path
	 * @return boolean
	 */
	public function removeFile( $remoteFile ) {
		$remoteFile = $this->sanitizePath( $remoteFile );

		// Remove file
		if ( @ftp_delete( $this->getConnection(), $remoteFile ) ) {
			return true;
		}

		return false;
	}

	/**
	 * Test FTP connection
	 *
	 * @return boolean
	 */
	public function testConnection() {
		try {
			return (bool) $this->getConnection();
		} catch ( Exception $e ) {
			return false;
		}
	}

	/**
	 * Test FTP append file
	 *
	 * @param  string $remoteFile Path to remote file
	 * @return boolean
	 */
	public function testAppendFile( $remoteFile = '.servmask' ) {

		try {

			// Write temporary file
			if ( ( $localFile = fopen( 'php://temp', 'wb+' ) ) ) {

				// Write PHP OS
				fwrite( $localFile, PHP_OS );

				// Set file pointer
				rewind( $localFile );

				// Upload file
				if ( ! @ftp_fput( $this->getConnection(), $remoteFile, $localFile, FTP_BINARY ) ) {
					throw new Exception( 'FTP upload has failed.' );
				}

				// Close file handler
				fclose( $localFile );
			}

			// Write temporary file (append)
			if ( ( $localFile = fopen( 'php://temp', 'wb+' ) ) ) {

				// Write PHP OS
				fwrite( $localFile, PHP_OS );

				// Write PHP OS (use offset parameter on upload)
				fwrite( $localFile, PHP_OS );

				// Set file pointer
				rewind( $localFile );

				// Upload file (append)
				if ( ! @ftp_fput( $this->getConnection(), $remoteFile, $localFile, FTP_BINARY, strlen( PHP_OS ) ) ) {
					throw new Exception( 'FTP upload has failed.' );
				}

				// Close file handler
				fclose( $localFile );
			}

			// Get file
			if ( ( $remoteSize = @ftp_size( $this->getConnection(), $remoteFile ) ) ) {
				if ( $remoteSize === ( 2 * strlen( PHP_OS ) ) ) {
					return true;
				}
			}

		} catch ( Exception $e ) {
		}

		return false;
	}

	/**
	 * Get FTP connection (lazy loading)
	 *
	 * @return resource
	 */
	protected function getConnection() {
		if ( $this->handler === null ) {
			if ( $this->ssl ) {
				if ( ! ( $handler = @ftp_ssl_connect( $this->getHostname(), $this->getPort() ) ) ) {
					throw new Exception( 'Unable to connect to FTPS server.' );
				}
			} else {
				if ( ! ( $handler = @ftp_connect( $this->getHostname(), $this->getPort() ) ) ) {
					throw new Exception( 'Unable to connect to FTP server.' );
				}
			}

			// Login to FTP server
			if ( ! @ftp_login( $handler, $this->getUsername(), $this->getPassword() ) ) {
				throw new Exception( 'Unable to login to FTP server.' );
			}

			// Turn on FTP passive mode
			if ( $this->getPassive() ) {
				@ftp_pasv( $handler, true );
			}

			// Change FTP directory
			if ( $this->getDirectory() ) {
				if ( ! @ftp_chdir( $handler, $this->getDirectory() ) ) {
					throw new Exception( 'Unable to change FTP directory.' );
				}
			}

			// Set FTP handler
			$this->handler = $handler;
		}

		return $this->handler;
	}

	/**
	 * Sanitize FTP path
	 *
	 * @param  string $path FTP path
	 * @return string
	 */
	protected function sanitizePath( $path ) {
		return ltrim( preg_replace( '/[\\\\\/]+/', '/', $path ), '/' );
	}

	/**
	 * Close FTP connection
	 *
	 * @return void
	 */
	public function __destruct() {
		if ( $this->handler !== null ) {
			@ftp_close( $this->getConnection() );
		}
	}
}
