<?php
/**
 * Copyright (C) 2014-2017 ServMask Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ███████╗███████╗██████╗ ██╗   ██╗███╗   ███╗ █████╗ ███████╗██╗  ██╗
 * ██╔════╝██╔════╝██╔══██╗██║   ██║████╗ ████║██╔══██╗██╔════╝██║ ██╔╝
 * ███████╗█████╗  ██████╔╝██║   ██║██╔████╔██║███████║███████╗█████╔╝
 * ╚════██║██╔══╝  ██╔══██╗╚██╗ ██╔╝██║╚██╔╝██║██╔══██║╚════██║██╔═██╗
 * ███████║███████╗██║  ██║ ╚████╔╝ ██║ ╚═╝ ██║██║  ██║███████║██║  ██╗
 * ╚══════╝╚══════╝╚═╝  ╚═╝  ╚═══╝  ╚═╝     ╚═╝╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝
 */

class ServMaskSFTPClient
{
	const CHUNK_SIZE          = 4194304; // 4 MB

	/**
	 * Holds SFTP server hostname
	 *
	 * @access protected
	 * @var    string $hostname SFTP server hostname
	 */
	protected $hostname       = null;

	/**
	 * Holds SFTP client authentication
	 *
	 * @access protected
	 * @var    string $authentication SFTP client authentication
	 */
	protected $authentication = null;

	/**
	 * Holds SFTP client username
	 *
	 * @access protected
	 * @var    string $username SFTP client username
	 */
	protected $username       = null;

	/**
	 * Holds SFTP client password
	 *
	 * @access protected
	 * @var    string $password SFTP client password
	 */
	protected $password       = null;

	/**
	 * Holds SFTP client private key
	 *
	 * @access protected
	 * @var    string $key SFTP client private key
	 */
	protected $key            = null;

	/**
	 * Holds SFTP client passphrase
	 *
	 * @access protected
	 * @var    string $passphrase SFTP client passphrase
	 */
	protected $passphrase     = null;

	/**
	 * Holds SFTP server directory
	 *
	 * @access protected
	 * @var    string $directory SFTP server directory
	 */
	protected $directory      = null;

	/**
	 * Holds SFTP server port
	 *
	 * @access protected
	 * @var    integer $port SFTP server port
	 */
	protected $port           = null;

	/**
	 * Holds SFTP library provider
	 *
	 * @access private
	 * @var    string $provider SFTP library provider
	 */
	private $provider         = null;

	/**
	 * Constructs the ServMaskSFTPClient object
	 *
	 * @param string  $hostname       SFTP server hostname
	 * @param string  $authentication SFTP client authentication
	 * @param string  $username       SFTP client username
	 * @param string  $password       SFTP client password
	 * @param string  $key            SFTP client private key
	 * @param string  $passphrase     SFTP client passphrase
	 * @param string  $directory      SFTP server directory
	 * @param integer $port           SFTP server port
	 * @param object  $provider       SFTP library provider
	 */
	public function __construct( $hostname, $authentication, $username, $password, $key, $passphrase, $directory = '/', $port = 22, $provider = null ) {
		$this->hostname       = $hostname;
		$this->authentication = $authentication;
		$this->username       = $username;
		$this->password       = $password;
		$this->key            = $key;
		$this->passphrase     = $passphrase;
		$this->directory      = $directory;
		$this->port           = $port;
		$this->provider       = $provider;

		// Set provider
		if ( empty( $this->provider ) ) {
			$this->provider = new Net_SFTP( $this->hostname, $this->port );
		}

		// Set credentials
		if ( $this->authentication === 'key' ) {
			$rsa = new Crypt_RSA;
			$rsa->setPassword( $this->passphrase );
			$rsa->loadKey( $this->key );

			// Login to server
			if ( $this->provider->login( $this->username, $rsa ) ) {
				if ( $this->directory ) {
					if ( $this->provider->chdir( $this->directory ) === false ) {
						$this->provider->disconnect();
					}
				}
			}
		} else {
			// Login to server
			if ( $this->provider->login( $this->username, $this->password ) ) {
				if ( $this->directory ) {
					if ( $this->provider->chdir( $this->directory ) === false ) {
						$this->provider->disconnect();
					}
				}
			}
		}
	}

	/**
	 * Create folder
	 *
	 * @param  string  $path SFTP path
	 * @return boolean
	 */
	public function createFolder( $path ) {
		return $this->provider->mkdir( $path );
	}

	/**
	 * List metadata of the items in the folder
	 *
	 * @param  string $path Path to be listed
	 * @return array
	 */
	public function listFolder( $path ) {
		// List raw files and directories
		if ( empty( $path ) ) {
			$items = $this->rawListFolder( '.' );
		} else {
			$items = $this->rawListFolder( $path );
		}

		// Loop over files and folders
		$objects = array();
		foreach ( $items as $item ) {
			if ( strpos( $item['filename'], '.' ) !== 0 ) {
				$objects[] = array(
					'name'  => isset( $item['filename'] ) ? basename( $item['filename'] ) : null,
					'path'  => isset( $item['filename'] ) ? $this->sanitizePath( sprintf( '/%s/%s', $path, $item['filename'] ) ) : null,
					'date'  => isset( $item['mtime'] ) ? $item['mtime'] : null,
					'size'  => isset( $item['size'] ) ? size_format( $item['size'] ) : null,
					'bytes' => isset( $item['size'] ) ? $item['size'] : null,
					'type'  => isset( $item['type'] ) && $item['type'] === NET_SFTP_TYPE_DIRECTORY ? 'folder' : 'file',
				);
			}
		}

		return $objects;
	}

	/**
	 * Retrieve files and folders raw metadata
	 *
	 * @param  string $path SFTP path
	 * @return array
	 */
	public function rawListFolder( $path ) {
		return $this->provider->rawlist( $path );
	}

	/**
	 * Upload file
	 *
	 * @param  string   $remoteFile SFTP path
	 * @param  resource $localFile  Local file
	 * @param  array    $params     File parameters
	 * @return array
	 */
	public function uploadFile( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );
		$localFileSize = filesize( $localFileMeta['uri'] );

		$this->provider->put( $remoteFile, $localFile, NET_SFTP_LOCAL_FILE );

		$params['offset'] = $localFileSize;

		return $params;
	}

	/**
	 * Upload file chunk
	 *
	 * @param  string   $remoteFile SFTP path
	 * @param  resource $localFile  Local file data
	 * @param  array    $params     File parameters
	 * @return array
	 */
	public function uploadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );
		$localFileSize = filesize( $localFileMeta['uri'] );

		$chunk = fopen( 'php://temp', 'wb+' );

		// Get chunk size
		if ( $localFileSize < ( $params['offset'] + self::CHUNK_SIZE ) ) {
			$chunkSize = $localFileSize - $params['offset'];
		} else {
			$chunkSize = self::CHUNK_SIZE;
		}

		// Read file chunk
		if ( ( $size = stream_copy_to_stream( $localFile, $chunk, $chunkSize, $params['offset'] ) ) ) {
			// Set chunk pointer
			rewind( $chunk );

			$this->provider->put( $remoteFile, $chunk, NET_SFTP_LOCAL_FILE | NET_SFTP_RESUME );

			$params['offset'] += $size;
		}

		return $params;
	}

	/**
	 * Download file
	 *
	 * @param  resource $remoteFile SFTP path
	 * @param  stream   $localFile  Local file
	 * @param  params   $params     File parameters
	 * @return array
	 */
	public function downloadFile( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );

		$this->provider->get( $remoteFile, $localFile );

		$params['offset'] = filesize( $localFileMeta['uri'] );

		return $params;
	}

	/**
	 * Download file chunk
	 *
	 * @param  resource $remoteFile SFTP path
	 * @param  stream   $localFile  Local file
	 * @param  array    $params     File parameters
	 * @return array
	 */
	public function downloadFileChunk( $remoteFile, $localFile, &$params = array() ) {
		$localFileMeta = stream_get_meta_data( $localFile );

		$this->provider->get( $remoteFile, $localFile, $params['offset'], self::CHUNK_SIZE );

		$params['offset'] = filesize( $localFileMeta['uri'] );

		return $params;
	}

	/**
	 * Remove file
	 *
	 * @param  string  $remoteFile SFTP path
	 * @return boolean
	 */
	public function removeFile( $remoteFile ) {
		return $this->provider->delete( $remoteFile );
	}

	/**
	 * Test SFTP connection
	 *
	 * @return boolean
	 */
	public function testConnection() {
		try {
			return (bool) $this->provider->isConnected() && (bool) $this->provider->isAuthenticated();
		} catch ( Exception $e ) {
			return false;
		}
	}

	/**
	 * Test SFTP append file
	 *
	 * @param  string $remoteFile Path to remote file
	 * @return boolean
	 */
	public function testAppendFile( $remoteFile = '.servmask' ) {

		try {

			// Write temporary file
			if ( ( $localFile = fopen( 'php://temp', 'w+' ) ) ) {

				// Write PHP OS
				fwrite( $localFile, PHP_OS );

				// Set file pointer
				rewind( $localFile );

				// Upload file
				if ( ! $this->provider->put( $remoteFile, $localFile, NET_SFTP_LOCAL_FILE ) ) {
					throw new Exception( 'SFTP upload has failed.' );
				}

				// Close file handler
				fclose( $localFile );
			}

			// Write temporary file (append)
			if ( ( $localFile = fopen( 'php://temp', 'wb+' ) ) ) {

				// Write PHP OS
				fwrite( $localFile, PHP_OS );

				// Set file pointer
				rewind( $localFile );

				// Upload file (append)
				if ( ! $this->provider->put( $remoteFile, $localFile, NET_SFTP_LOCAL_FILE | NET_SFTP_RESUME ) ) {
					throw new Exception( 'SFTP upload has failed.' );
				}

				// Close file handler
				fclose( $localFile );
			}

			$this->provider->clearStatCache();

			// Get file
			if ( ( $remoteSize = $this->provider->size( $remoteFile ) ) ) {
				if ( $remoteSize === ( 2 * strlen( PHP_OS ) ) ) {
					return true;
				}
			}

		} catch ( Exception $e ) {
		}

		return false;
	}

	/**
	 * Sanitize SFTP path
	 *
	 * @param  string $path SFTP path
	 * @return string
	 */
	protected function sanitizePath( $path ) {
		return ltrim( preg_replace( '/[\\\\\/]+/', '/', $path ), '/' );
	}
}
